/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl;

import java.io.File;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.Factory;
import org.objectweb.fractal.adl.FactoryFactory;
import org.objectweb.fractal.adl.JavaFactory;
import org.objectweb.fractal.adl.StaticJavaGenerator;
import org.objectweb.fractal.adl.util.Printer;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.cecilia.adl.plugin.AbstractPluginManager;
import org.objectweb.fractal.util.Fractal;

/**
 * A class to compile a component. Usage: Launcher &lt;definition&gt; where
 * &lt;definition&gt; is the name of the component to be compiled.
 */
public class Launcher extends AbstractLauncher {

  /**
   * The name of the {@link System#getProperty(String) system property} that
   * specify the backend to use for the plugin factory.
   */
  public static final String PLUGIN_FACTORY_BACKEND_PROPERTY_NAME = AbstractPluginManager.PLUGIN_FACTORY_BACKEND;

  /**
   * The default value of the {@link #PLUGIN_FACTORY_BACKEND_PROPERTY_NAME}
   * property.
   */
  public static final String DEFAULT_PLUGIN_FACTORY_BACKEND       = "Java";

  /** The name of the default ADL of the compiler. */
  public static final String CECILIA_BASIC_COMPILER_ADL           = "org.objectweb.fractal.cecilia.adl.CeciliaBasicFactory";

  private final CmdOption    compilerADLOpt                       = new CmdOption(
                                                                      new String[]{"compiler-adl"},
                                                                      "Specify the name of the ADL of the compiler itself",
                                                                      CECILIA_BASIC_COMPILER_ADL);

  private final CmdOption    compilerCmdOpt                       = new CmdOption(
                                                                      new String[]{
      "compiler-command", "cc"                                        },
                                                                      "the command of the C compiler (default is 'gcc')",
                                                                      "gcc");

  private final CmdOption    defsOpt                              = new CmdAppendOption(
                                                                      new String[]{
      "defs", "D"                                                     },
                                                                      "the -D compiler directives");
  private final CmdOption    cFlagsOpt                            = new CmdAppendOption(
                                                                      new String[]{
      "c-flags", "cf"                                                 },
                                                                      "the c-flags compiler directives");

  private final CmdOption    includePathOpt                       = new CmdPathOption(
                                                                      new String[]{
      "inc-path", "I"                                                 },
                                                                      "the -I compiler directives");

  private final CmdOption    linkerCmdOpt                         = new CmdOption(
                                                                      new String[]{
      "linker-command", "ld"                                          },
                                                                      "the command of the linker (default is 'gcc')",
                                                                      "gcc");

  private final CmdOption    executableFormat                     = new CmdOption(
                                                                      new String[]{"executable-format"},
                                                                      "the binary format associated to \"executable\" output format (default is 'elf')",
                                                                      "elf");

  private final CmdOption    ldFlagsOpt                           = new CmdAppendOption(
                                                                      new String[]{
      "ld-flags", "lf"                                                },
                                                                      "the ld-flags compiler directives");
  private final CmdOption    ldPathOpt                            = new CmdPathOption(
                                                                      new String[]{
      "ld-path", "-L"                                                 },
                                                                      "the -L compiler directives");

  private final CmdOption    archiveCmdOpt                        = new CmdOption(
                                                                      new String[]{
      "archive-command", "ar"                                         },
                                                                      "the command of the archive tool, (default is 'ar')",
                                                                      "ar");

  private final CmdOption    linkerScriptOpt                      = new CmdOption(
                                                                      new String[]{
      "linker-script", "T"                                            },
                                                                      "linker script");

  private final CmdOption    executableNameCmdOpt                 = new CmdOption(
                                                                      new String[]{
      "executable-name", "exec-name"                                  },
                                                                      "the name of the target executable file (can be used with only one ADL to be compiled.)",
                                                                      "");

  private final CmdOption    concurrentJobCmdOpt                  = new CmdOption(
                                                                      new String[]{
      "jobs", "j"                                                     },
                                                                      "The number of concurrent compilation jobs",
                                                                      "1");

  Map<String, Object>        compilerContext                      = new HashMap<String, Object>();
  Factory                    compiler;

  /**
   * Creates a new Compiler launcher context.
   * 
   * @param outDir the output directory
   * @param sources the source path.
   */
  public Launcher(final String outDir, final String sources) {
    this(new String[]{"-o=" + outDir, "-src-path=" + sources,
        "this_is_not_an_adl"}, false);
  }

  /**
   * Creates a new Compiler launcher context.
   * 
   * @param options the options
   */
  public Launcher(final Map<String, String> options) {
    this(toArgs(options), false);
  }

  private static String[] toArgs(final Map<String, String> options) {
    final String[] params = new String[options.size()];
    int i = 0;
    for (final Map.Entry<String, String> option : options.entrySet()) {
      params[i] = option.getKey() + '=' + option.getValue();
      i++;
    }
    return params;
  }

  /**
   * Create the o.o.f.a.Factory class that will be used to read the application
   * definition. Subclasses of the Launcher may extend this.
   * 
   * @param compilerDefinition
   * @param compilerContext
   * @return
   * @throws ADLException
   * @throws Exception
   * @throws Exception
   */
  protected Factory createCompiler(final Map<String, Object> compilerContext) {
    final String compilerAdlDefinition = compilerADLOpt.getValue();

    final Object pluginFactoryBackend = compilerContext
        .get(PLUGIN_FACTORY_BACKEND_PROPERTY_NAME);
    if (pluginFactoryBackend == null
        || DEFAULT_PLUGIN_FACTORY_BACKEND.equals(pluginFactoryBackend)) {
      /* instantiate the bootstrap factory with the default backend */
      try {
        final Class<?> factoryFactoryClass = getClass().getClassLoader()
            .loadClass(
                compilerAdlDefinition + StaticJavaGenerator.CLASSNAME_SUFFIX);
        final JavaFactory factoryFactory = (JavaFactory) factoryFactoryClass
            .newInstance();
        final Map<?, ?> compilerComponent = (Map<?, ?>) factoryFactory
            .newComponent();
        Printer
            .debug("[Launcher] Cecilia compiler instantiated using statically generated factory.");
        return (Factory) compilerComponent.get("factory");
      } catch (final Exception e) {
        // ignore and try to instantiate the compiler using a FractalADL
        // factory.
        if (Printer.isDebugEnable())
          Printer
              .debug("[Launcher] Unable to instantiate cecilia compiler using statically generated factory ("
                  + e.getClass().getName() + ":" + e.getMessage() + ").");
      }

      try {
        final Factory factory = FactoryFactory.getFactory();
        /* try to instantiate the CeciliaADL BasicFactory */
        final Map<?, ?> compilerComponent = (Map<?, ?>) factory.newComponent(
            compilerAdlDefinition, compilerContext);
        Printer
            .debug("[Launcher] Cecilia compiler instantiated using FractalADL factory.");
        return (Factory) compilerComponent.get("factory");
      } catch (final ADLException e) {
        internalError("Unable to instantiate Cecilia compiler", e, 101);
      }

    } else {
      /* instantiate the bootstrap factory with the Fractal backend */
      Component compilerComponent = null;
      try {
        final Factory factory = FactoryFactory
            .getFactory(FactoryFactory.FRACTAL_BACKEND);

        /* try to instantiate the CeciliaADL BasicFactory */
        compilerComponent = (Component) factory.newComponent(
            compilerAdlDefinition, compilerContext);
        /* start compiler */
        Fractal.getLifeCycleController(compilerComponent).startFc();

        final Factory result = (Factory) compilerComponent
            .getFcInterface("factory");
        return result;
      } catch (final ADLException e) {
        internalError("Unable to instantiate Cecilia compiler", e, 101);
      } catch (final NoSuchInterfaceException e) {
        internalError("Unable to instantiate Cecilia compiler", e, 101);
      } catch (final IllegalLifeCycleException e) {
        internalError("Unable to instantiate Cecilia compiler", e, 101);
      }
    }
    // unreachable code since internalError never exit.
    return null;
  }

  protected Launcher(final String[] args, final boolean exitOnFailure) {
    super(exitOnFailure);

    if (Printer.isDebugEnable()) {
      for (final String arg : args) {
        Printer.debug("[arg] " + arg);
      }
    }

    setSpecificOptions(new CmdOption[]{compilerADLOpt, compilerCmdOpt, defsOpt,
        cFlagsOpt, includePathOpt, linkerCmdOpt, ldFlagsOpt, ldPathOpt,
        executableFormat, archiveCmdOpt, linkerScriptOpt, executableNameCmdOpt,
        concurrentJobCmdOpt});

    parseArgs(args);

    /*
     * get the CeciliaADL Compiler (default implementation a pure Java
     * component)
     */

    this.compiler = createCompiler(compilerContext);

    final List<String> incPaths = new ArrayList<String>(sourcePaths);
    if (this.includePathOpt.getValue() != null)
      incPaths.addAll(parsePathList(includePathOpt.getValue()));
    StringBuffer sb = new StringBuffer();
    for (final String inc : incPaths) {
      sb.append(" -I").append(inc);
    }
    final String includePath = sb.toString();

    final List<String> ldPaths = new ArrayList<String>();
    if (ldPathOpt.getValue() != null)
      ldPaths.addAll(parsePathList(ldPathOpt.getValue()));
    sb = new StringBuffer();
    for (final String ld : ldPaths) {
      sb.append(" -L").append(ld);
    }
    final String ldPath = sb.toString();

    // add compiler arguments to context
    compilerContext.put(compilerCmdOpt.options[0], compilerCmdOpt.getValue());
    compilerContext.put(cFlagsOpt.options[0], cFlagsOpt.getValue());
    compilerContext.put(includePathOpt.options[0], includePath);
    compilerContext.put(defsOpt.options[0], defsOpt.getValue());

    compilerContext.put(linkerCmdOpt.options[0], linkerCmdOpt.getValue());
    compilerContext.put(ldFlagsOpt.options[0], ldFlagsOpt.getValue());
    compilerContext.put(ldPathOpt.options[0], ldPath);
    compilerContext.put(archiveCmdOpt.options[0], archiveCmdOpt.getValue());
    compilerContext.put(executableFormat.options[0], executableFormat
        .getValue());

    Integer jobs = null;
    try {
      jobs = Integer.decode(concurrentJobCmdOpt.getValue());
    } catch (final NumberFormatException e) {
      error("Invalid jobs value '" + concurrentJobCmdOpt.getValue()
          + "' is not a valid number", 1);
    }
    compilerContext.put(concurrentJobCmdOpt.options[0], jobs);

    // add build directories to context
    compilerContext.put("adlBuildDirectory", adlBuildDir);
    compilerContext.put("idlBuildDirectory", idlBuildDir);
    compilerContext.put("objBuildDirectory", objBuildDir);

    // add source class loader in context
    compilerContext.put("classloader", sourceClassLoader);

    // add linker script to the context
    if (linkerScriptOpt.getValue() != null)
      compilerContext.put(linkerScriptOpt.options[0], linkerScriptOpt
          .getValue());

  }

  /**
   * Compiles the given ADL.
   * 
   * @param adl the name of the the name of the architecture (ADL) to compile.
   * @return the list of objects that are the results of the compilation.
   * @throws ADLException if the compilation fails.
   */
  public List<Object> compile(final String adlName) throws ADLException {
    adlNameMap = new TreeMap<String, String>();
    adlNameMap.put(adlName, null);
    return compile();
  }

  /**
   * Compiles the given ADL.
   * 
   * @param adlName the name of the architecture (ADL) to compile.
   * @param execName the name of the executable file to be created.
   * @return the list of objects that are the results of the compilation.
   * @throws ADLException if the compilation fails.
   */
  public List<Object> compile(final String adlName, final String execName)
      throws ADLException {
    adlNameMap = new TreeMap<String, String>();
    adlNameMap.put(adlName, executableNameCmdOpt.defaultValue);
    return compile();
  }

  /**
   * Compiles the list of adl files that are received from the command-line.
   * 
   * @return the list of objects that are the results of the compilation of the
   *         list of adl files received from the command-line.
   * @throws ADLException if the compilation fails.
   */
  private List<Object> compile() throws ADLException {
    final List<Object> result = new ArrayList<Object>();
    for (final String adlName : adlNameMap.keySet()) {
      compilerContext.put(executableNameCmdOpt.options[0], adlNameMap
          .get(adlName));
      final Object compilationResult = this.compiler.newComponent(adlName,
          new HashMap<Object, Object>(compilerContext));
      if (compilationResult instanceof File)
        result.add(compilationResult);
      else if (compilationResult instanceof Collection)
        for (final Object elem : (Collection<?>) compilationResult) {
          result.add(elem);
        }
    }
    return result;
  }

  @Override
  protected void printUsage(final PrintStream ps) {
    ps.println("Usage: Launcher [OPTIONS] (<definition>[:<execname>])+");
    ps.println("  where <definition> is the name of the component to"
        + " be compiled, ");
    ps
        .println("  and <execname> is the name of the output file to be created.");
  }

  /**
   * Entry point.
   * 
   * @param args
   */
  public static void main(final String[] args) {
    doMain(new Launcher(args, true));
  }

  protected static void doMain(final Launcher launcher) {
    try {
      launcher.compile();
    } catch (final ADLException e) {
      e.printStackTrace();
// if (Printer.isDebugEnable()) {
// e.printStackTrace();
// } else {
// if (e instanceof TaskExecutionException) {
// final TaskExecutionException tee = (TaskExecutionException) e;
// final Throwable cause = tee.getCause();
// if (cause instanceof CompilationException) System.exit(-1);
// }

      final StringBuffer sb = new StringBuffer();
      sb.append(e.getMessage()).append('\n');
      Throwable cause = e.getCause();
      while (cause != null) {
        sb.append("caused by : ");
        if (cause instanceof ClassNotFoundException) {
          sb.append("Can't find IDL, ADL or source file: '").append(
              cause.getMessage()).append("'");
        } else if (cause instanceof ADLException) {
          sb.append(cause.getMessage());
        } else {
          final StringWriter sw = new StringWriter();
          cause.printStackTrace(new PrintWriter(sw));
          sb.append(sw.getBuffer());
        }
        sb.append('\n');
        cause = cause.getCause();
      }
      System.err.println(sb);
      System.exit(-1);
    }

  }

}
