/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2008 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author:Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl.compiler;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;

import org.objectweb.fractal.adl.util.Printer;
import org.objectweb.fractal.cecilia.adl.file.FutureFileCollectionProvider;
import org.objectweb.fractal.cecilia.adl.file.FutureFileProvider;
import org.objectweb.fractal.task.core.primitive.annotations.ClientInterface;
import org.objectweb.fractal.task.core.primitive.annotations.ServerInterface;
import org.objectweb.fractal.task.core.primitive.annotations.ServerInterfaces;
import org.objectweb.fractal.task.core.primitive.annotations.TaskParameters;

@TaskParameters("id")
@ServerInterfaces(@ServerInterface(name = "file-provider", signature = FutureFileProvider.class, record = "role:linkedFile, id:%", parameters = "id"))
public abstract class AbstractArchiveTask implements FutureFileProvider {

  /** The name of the linked file. */
  protected final String                                 outputName;
  /** The compilation flags. */
  protected String                                       flags;
  /** The output directory. */
  protected final File                                   outputDir;

  // Produced file
  protected Future<File>                                 archiveFile;

  // ---------------------------------------------------------------------------
  // Task client interfaces
  // ---------------------------------------------------------------------------

  /** The client collection interface used to retrieve the files to linked. */
  @ClientInterface(name = "source-file-provider", signature = FutureFileProvider.class, record = "role:compiledFile, id:%", parameters = "id")
  public final Map<String, FutureFileProvider>           fileProviderItfs           = new HashMap<String, FutureFileProvider>();

  /** The client collection interface used to retrieve the files to linked. */
  @ClientInterface(name = "source-file-collection-provider", signature = FutureFileCollectionProvider.class, record = "role:compiledFileCollection, id:%", parameters = "id")
  public final Map<String, FutureFileCollectionProvider> fileCollectionProviderItfs = new HashMap<String, FutureFileCollectionProvider>();

  // ---------------------------------------------------------------------------
  // Task constructor
  // ---------------------------------------------------------------------------

  /**
   * @param outputName the name of the linked file
   * @param outputDir the directory into which the compiled file will be placed.
   * @param flags compilation flags.
   */
  public AbstractArchiveTask(final String outputName, final File outputDir,
      final String flags) {
    this.outputName = outputName;
    this.outputDir = outputDir;
    this.flags = flags;
  }

  // ---------------------------------------------------------------------------
  // Abstract methods
  // ---------------------------------------------------------------------------

  /**
   * Archive method.
   * 
   * @param outputFile the linked file that is provided by this task.
   * @param inputFiles the input files to be linked.
   * @throws the archive file. Most of the time this is the same the
   *             <code>outputFile</code> parameter but it is not necessary the
   *             case.
   * @throws Exception if an error occurs.
   */
  protected abstract File archivate(File outputFile, Collection<File> inputFiles)
      throws Exception;

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  protected File getOutputFile() {
    return new File(outputDir, outputName);
  }

  protected void createParentDir(final File outputFile)
      throws CompilationException {
    final File parent = outputFile.getParentFile();
    if (parent.exists()) {
      if (!parent.isDirectory()) {
        throw new CompilationException("Invalid output file '" + outputFile
            + "': the parent file is not a directory");
      }
    } else {
      if (!parent.mkdirs()) {
        throw new CompilationException(
            "Can't create parent directory of file '" + outputFile + "'");
      }
    }
  }

  // -------------------------------------------------------------------------
  // Implementation of the FileProvider interface
  // -------------------------------------------------------------------------

  public synchronized Future<File> getFile(final ExecutorService executorService) {
    if (archiveFile == null) {
      archiveFile = executorService.submit(new Callable<File>() {

        public File call() throws Exception {
          final Collection<Future<File>> futureFiles = new HashSet<Future<File>>();

          // get every files provided by every FileProvider
          for (final FutureFileProvider fileProvider : fileProviderItfs
              .values()) {
            futureFiles.add(fileProvider.getFile(executorService));
          }

          // get every files provided by every FileCollectionProvider
          for (final FutureFileCollectionProvider collectionProvider : fileCollectionProviderItfs
              .values()) {
            futureFiles.addAll(collectionProvider.getFiles(executorService));
          }

          final Collection<File> inputFiles = new ArrayList<File>(futureFiles
              .size());
          File archiveFile = getOutputFile();
          final long outputTimestamp = archiveFile.lastModified();
          boolean relink;
          if (outputTimestamp > 0) {
            relink = false;
            for (final Future<File> futureFile : futureFiles) {
              final File inputFile = futureFile.get();
              inputFiles.add(inputFile);
              if (inputFile.lastModified() > outputTimestamp) {
                if (Printer.isDebugEnable())
                  Printer.debug("Input file '" + inputFile
                      + " is more recent than file '" + archiveFile
                      + "', recompile.");
                relink = true;
                break;
              }
            }
            if (!relink && Printer.isDebugEnable())
              Printer.debug("Output file '" + archiveFile
                  + "' is up to date, do not recompile.");
          } else {
            relink = true;
          }

          if (relink) {
            createParentDir(archiveFile);
            archiveFile = archivate(archiveFile, inputFiles);
          }
          return archiveFile;
        }
      });
    }
    return archiveFile;
  }
}
