/***
 * Cecilia ADL Compiler
 * Copyright (C) 2008 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author:Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl.compiler.gnu;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;

import org.objectweb.fractal.adl.util.Printer;
import org.objectweb.fractal.task.codegen.util.SourceFile;

public class GnuDepCompilerTask extends GnuCompilerTask {

  /** The default file extension of dependency files. */
  protected static final String DEPENDENCY_EXTENSION = ".d";

  // ---------------------------------------------------------------------------
  // Task constructor
  // ---------------------------------------------------------------------------

  /**
   * @param gccCommand the command of the Gnu C compiler to be executed by this
   *            task.
   * @param outputDir the directory into which the compiled file will be placed.
   * @param flags compilation flags.
   */
  public GnuDepCompilerTask(final String gccCommand, final File outputDir,
      final String flags) {
    super(gccCommand, outputDir, flags);
  }

  // ---------------------------------------------------------------------------
  // Overridden methods
  // ---------------------------------------------------------------------------

  @Override
  protected void compile(final SourceFile sourceFile, final File outputFile)
      throws Exception {
    final File depFile = getDependencyFile(outputFile);

    if (needRecompile(outputFile, depFile)) {
      flags = "-MMD -MF " + depFile.getCanonicalPath() + " " + flags;
      super.compile(sourceFile, outputFile);
    }
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  protected File getDependencyFile(final File outputFile) {
    String name = outputFile.getName();
    name = name.substring(0, name.lastIndexOf('.')) + DEPENDENCY_EXTENSION;
    return new File(outputFile.getParent(), name);
  }

  protected boolean needRecompile(final File outputFile,
      final File dependencyFile) throws IOException {

    if (!outputFile.exists()) {
      if (Printer.isDebugEnable())
        Printer.debug("Output file '" + outputFile
            + "' does not exist, compile it.");
      return true;
    }
    final long outputTimestamp = outputFile.lastModified();
    if (outputTimestamp == 0) {
      if (Printer.isWarningEnable())
        Printer.warning("Unable to determine timestamp of file '" + outputFile
            + "', recompile.");
      return true;
    }

    if (!dependencyFile.exists()) {
      if (Printer.isDebugEnable())
        Printer.debug("Dependency file of '" + outputFile
            + "' does not exist, recompile.");
      return true;
    }

    // Read entire file
    final BufferedReader reader = new BufferedReader(new FileReader(
        dependencyFile));
    String line = null, dependencies = "";
    while ((line = reader.readLine()) != null) {
      dependencies += line + "\n";
    }

    final String[] tokens = dependencies.split("(\\s*\\\\\\n\\s*)|\\s+");

    // look for the first non empty token
    int i = 0;
    while (i < tokens.length && tokens[i].length() == 0)
      i++;

    if (i == tokens.length
        || !tokens[i].startsWith(outputFile.getCanonicalPath())) {
      // the dependency file doesn't seem to be valid, recompile
      Printer.debug("Invalid dependency file " + dependencyFile);
      return true;
    }

    i++;
    for (; i < tokens.length; i++) {
      if (tokens[i].length() == 0) continue;

      final File f = new File(tokens[i]);
      if (!f.exists()) {
        if (Printer.isDebugEnable())
          Printer.debug("Warning input file '" + f + " does not exist.");
      } else if (f.lastModified() > outputTimestamp) {
        if (Printer.isDebugEnable())
          Printer.debug("Input file '" + f
              + " is more recent than output file '" + outputFile
              + "', recompile.");
        return true;
      }
    }
    if (Printer.isDebugEnable())
      Printer.debug("Output file '" + outputFile
          + "' is up to date, do not recompile.");
    return false;
  }
}
