/***
 * Fractal ADL Parser
 * Copyright (C) 2002-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Eric Bruneton
 *
 * Contributor(s): Philippe Merle
 *                 Matthieu Leclercq (externalize interface and implementation
 *   loading)
 */

package org.objectweb.fractal.cecilia.adl.implementations;

import static org.objectweb.fractal.cecilia.adl.implementations.ImplementationDecorationUtil.setCode;

import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.AbstractLoader;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.implementations.Controller;
import org.objectweb.fractal.adl.implementations.ControllerContainer;
import org.objectweb.fractal.adl.implementations.Implementation;
import org.objectweb.fractal.adl.implementations.ImplementationCodeLoader;
import org.objectweb.fractal.adl.implementations.ImplementationContainer;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.cecilia.adl.directives.Include;
import org.objectweb.fractal.cecilia.adl.directives.IncludeContainer;

/**
 * A {@link org.objectweb.fractal.adl.Loader} to check {@link Implementation}
 * nodes in definitions.
 */
public class ImplementationLoader extends AbstractLoader {

  // --------------------------------------------------------------------------
  // Client interfaces
  // --------------------------------------------------------------------------

  /**
   * The name of the {@link ImplementationCodeLoader} client interface used by
   * this loader to load implementation code.
   */
  public static final String         IMPLEMENTATION_LOADER_ITF = "implementation-code-loader";

  /**
   * The {@link ImplementationCodeLoader} used by this loader to load
   * implementation code.
   */
  protected ImplementationCodeLoader implementationCodeLoaderItf;

  // --------------------------------------------------------------------------
  // Implementation of the Loader interface
  // --------------------------------------------------------------------------

  public Definition load(final String name, final Map<Object, Object> context)
      throws ADLException {
    final Definition d = clientLoader.load(name, context);
    checkNode(d, context);
    return d;
  }

  // --------------------------------------------------------------------------
  // Checking methods
  // --------------------------------------------------------------------------

  protected void checkNode(final Object node, final Map<Object, Object> context)
      throws ADLException {
    if (node instanceof ImplementationContainer) {
      checkImplementationContainer((ImplementationContainer) node, context);
    }
    if (node instanceof ControllerContainer) {
      checkControllerContainer((ControllerContainer) node);
    }
    if (node instanceof ComponentContainer) {
      for (final Component comp : ((ComponentContainer) node).getComponents()) {
        checkNode(comp, context);
      }
    }
  }

  protected void checkImplementationContainer(
      final ImplementationContainer container, final Map<Object, Object> context)
      throws ADLException {
    final Implementation impl = container.getImplementation();
    if (impl == null) return;

    if (!(impl instanceof ExtendedImplementation))
      throw new ADLException("Invalid implementation node. Must implements the"
          + " ExtendedImplementation interface. Check the DTD", impl);

    final ExtendedImplementation extImpl = (ExtendedImplementation) impl;
    final String className = impl.getClassName();
    if (className == null)
      throw new ADLException("Implementation class name missing", impl);

    final String language = extImpl.getLanguage();

    setCode(impl, implementationCodeLoaderItf.loadImplementation(className,
        language, context));

    if (impl instanceof IncludeContainer) {
      for (final Include include : ((IncludeContainer) impl).getIncludes()) {
        setCode((Node) include, implementationCodeLoaderItf.loadImplementation(
            include.getFile(), language, context));
      }
    }
  }

  protected void checkControllerContainer(final ControllerContainer container)
      throws ADLException {
    final Controller ctrl = container.getController();
    if (ctrl != null) {
      if (ctrl.getDescriptor() == null) {
        throw new ADLException("Controller descriptor missing", ctrl);
      }
    }
  }

  // --------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // --------------------------------------------------------------------------

  @Override
  public String[] listFc() {
    final String[] superList = super.listFc();
    final String[] list = new String[superList.length + 1];
    System.arraycopy(superList, 0, list, 0, superList.length);
    list[superList.length] = IMPLEMENTATION_LOADER_ITF;
    return list;
  }

  @Override
  public void bindFc(final String s, final Object o)
      throws NoSuchInterfaceException, IllegalBindingException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(IMPLEMENTATION_LOADER_ITF)) {
      this.implementationCodeLoaderItf = (ImplementationCodeLoader) o;
    } else {
      super.bindFc(s, o);
    }
  }

  @Override
  public Object lookupFc(final String s) throws NoSuchInterfaceException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(IMPLEMENTATION_LOADER_ITF)) {
      return this.implementationCodeLoaderItf;
    } else {
      return super.lookupFc(s);
    }
  }

  @Override
  public void unbindFc(final String s) throws IllegalBindingException,
      NoSuchInterfaceException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(IMPLEMENTATION_LOADER_ITF)) {
      this.implementationCodeLoaderItf = null;
    } else {
      super.unbindFc(s);
    }
  }
}
