/***
 * Fractal Task Codegen Framework
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.task.codegen.lib;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;

import org.objectweb.fractal.task.codegen.api.FactoryProviderTask;
import org.objectweb.fractal.task.codegen.api.FileCollectionProviderTask;
import org.objectweb.fractal.task.codegen.api.FileProviderTask;
import org.objectweb.fractal.task.codegen.api.RequireFactoryProviderTask;
import org.objectweb.fractal.task.codegen.api.RequireFileCollectionProviderTask;
import org.objectweb.fractal.task.codegen.api.RequireFileProviderTask;
import org.objectweb.fractal.task.core.AbstractTask;
import org.objectweb.fractal.task.core.TaskMap;

/**
 * Abstract implementation of {@link RequireFileCollectionProviderTask} task
 * interface. This class also implements the {@link RequireFactoryProviderTask}
 * interface so concrete task may depends on a
 * {@link FactoryProviderTask factory provider task}.
 */
public abstract class AbstractRequireFileCollectionProviderTask
    extends
      AbstractTask
    implements
      RequireFileCollectionProviderTask,
      RequireFileProviderTask,
      RequireFactoryProviderTask {

  private TaskMap.TaskHole                   factoryProviderTask;
  private final Collection<TaskMap.TaskHole> fileCollectionProviderTasks = new LinkedHashSet<TaskMap.TaskHole>();
  private final Collection<TaskMap.TaskHole> fileProviderTasks           = new LinkedHashSet<TaskMap.TaskHole>();

  // --------------------------------------------------------------------------
  // Utility method
  // --------------------------------------------------------------------------

  /**
   * Returns a collection of files provided by
   * {@link FileCollectionProviderTask file collection providers} and
   * {@link FileProviderTask file providers} this task requires.
   */
  protected Collection<File> getProvidedFiles() {
    final Collection<File> files = new HashSet<File>();
    for (final TaskMap.TaskHole collectionProvider : fileCollectionProviderTasks)
      for (final File file : ((FileCollectionProviderTask) collectionProvider
          .getTask()).getFiles())
        files.add(file);

    for (final TaskMap.TaskHole singleProvider : fileProviderTasks)
      files.add(((FileProviderTask) singleProvider.getTask()).getFile());

    return files;
  }

  // --------------------------------------------------------------------------
  // Implementation of the RequireFactoryProviderTask interface
  // --------------------------------------------------------------------------

  public FactoryProviderTask getFactoryProviderTask() {
    return (factoryProviderTask == null)
        ? null
        : (FactoryProviderTask) factoryProviderTask.getTask();
  }

  public void setFactoryProviderTask(final TaskMap.TaskHole task) {
    if (factoryProviderTask != null) removePreviousTask(factoryProviderTask);
    factoryProviderTask = task;
    if (factoryProviderTask != null) addPreviousTask(factoryProviderTask);
  }

  // --------------------------------------------------------------------------
  // Implementation of the RequireFileCollectionProviderTask interface
  // --------------------------------------------------------------------------

  public void addFileCollectionProviderTask(final TaskMap.TaskHole task) {
    if (fileCollectionProviderTasks.add(task)) addPreviousTask(task);
  }

  public void removeFileCollectionProviderTask(final TaskMap.TaskHole task) {
    if (fileCollectionProviderTasks.remove(task)) removePreviousTask(task);
  }

  public Collection<FileCollectionProviderTask> getFileCollectionProviderTasks() {
    final Collection<FileCollectionProviderTask> col = new ArrayList<FileCollectionProviderTask>();
    for (final TaskMap.TaskHole task : fileCollectionProviderTasks) {
      col.add((FileCollectionProviderTask) task.getTask());
    }
    return col;
  }

  // --------------------------------------------------------------------------
  // Implementation of the RequireFileProviderTask interface
  // --------------------------------------------------------------------------

  public void addFileProviderTask(final TaskMap.TaskHole task) {
    if (fileProviderTasks.add(task)) addPreviousTask(task);
  }

  public FileProviderTask[] getFileProviderTasks() {
    final FileProviderTask[] tasks = new FileProviderTask[fileProviderTasks
        .size()];
    int i = 0;
    for (final TaskMap.TaskHole hole : fileProviderTasks) {
      tasks[i++] = (FileProviderTask) hole.getTask();
    }
    return tasks;
  }

  public void removeFileProviderTask(final TaskMap.TaskHole task) {
    if (fileProviderTasks.remove(task)) removePreviousTask(task);
  }

  // --------------------------------------------------------------------------
  // Overridden AbstractTask methods
  // --------------------------------------------------------------------------

  @Override
  protected boolean doAddDependency(final TaskMap.TaskHole task,
      final Class<?> role, final Object context) {
    if (super.doAddDependency(task, role, context)) return true;

    if (role == RequireFileCollectionProviderTask.FILE_COLLECTION_PROVIDER_TASK_ROLE)
      addFileCollectionProviderTask(task);
    else if (role == RequireFileProviderTask.FILE_PROVIDER_TASK_ROLE)
      addFileProviderTask(task);
    else if (role == RequireFactoryProviderTask.FACTORY_PROVIDER_TASK_ROLE)
      setFactoryProviderTask(task);
    else
      return false;

    return true;
  }

  @Override
  protected boolean doRemoveDependency(final TaskMap.TaskHole task,
      final Class<?> role) {
    if (super.doRemoveDependency(task, role)) return true;

    if (role == RequireFileCollectionProviderTask.FILE_COLLECTION_PROVIDER_TASK_ROLE)
      removeFileCollectionProviderTask(task);
    else if (role == RequireFileProviderTask.FILE_PROVIDER_TASK_ROLE)
      removeFileProviderTask(task);
    else if (role == RequireFactoryProviderTask.FACTORY_PROVIDER_TASK_ROLE)
      setFactoryProviderTask(null);
    else
      return false;

    return true;
  }
}
