/***
 * Fractal Task Codegen Framework
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.task.codegen.lib;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.TreeMap;

import org.objectweb.fractal.task.codegen.api.FactoryProviderTask;
import org.objectweb.fractal.task.codegen.api.InstanceProviderTask;
import org.objectweb.fractal.task.codegen.api.RequireFactoryProviderTask;
import org.objectweb.fractal.task.codegen.api.RequireInstanceProviderTask;
import org.objectweb.fractal.task.codegen.api.RequireSourceCodeProviderTask;
import org.objectweb.fractal.task.codegen.api.SourceCodeProviderTask;
import org.objectweb.fractal.task.core.AbstractTask;
import org.objectweb.fractal.task.core.Task;
import org.objectweb.fractal.task.core.TaskMap;
import org.objectweb.fractal.task.core.TaskMap.TaskHole;

/**
 * Abstract implementation of {@link RequireSourceCodeProviderTask} task
 * interface.
 */
public abstract class AbstractRequireSourceCodeProviderTask
    extends
      AbstractTask
    implements
      RequireFactoryProviderTask,
      RequireInstanceProviderTask,
      RequireSourceCodeProviderTask {

  /**
   * Use this oder if no order is expected for the
   * {@link SourceCodeProviderTask source code providers} required by this task.
   */
  public static final int                     NO_ORDER          = 0;

  /**
   * Use this oder if the {@link SourceCodeProviderTask source code providers}
   * required by this task should be sorted according to the natural order of
   * their names.
   * 
   * @see TreeMap
   */
  public static final int                     NATURAL_KEY_ORDER = 1;

  /**
   * Use this oder if the {@link SourceCodeProviderTask source code providers}
   * required by this task should be sorted according to their insertion order.
   * 
   * @see LinkedHashMap
   */
  public static final int                     INSERTION_ORDER   = 2;

  private TaskMap.TaskHole                    factoryProviderTask;
  private TaskMap.TaskHole                    instanceProviderTask;
  private final Map<String, TaskMap.TaskHole> sourceCodeProviders;
  private final int                           order;

  protected AbstractRequireSourceCodeProviderTask(final int order) {
    this.order = order;
    sourceCodeProviders = newMap();
  }

  protected AbstractRequireSourceCodeProviderTask() {
    this(INSERTION_ORDER);
  }

  private <T> Map<String, T> newMap() {
    if (order == NO_ORDER) return new HashMap<String, T>();
    if (order == NATURAL_KEY_ORDER) return new TreeMap<String, T>();
    if (order == INSERTION_ORDER) return new LinkedHashMap<String, T>();
    throw new IllegalArgumentException("unknown order");
  }

  /**
   * Returns a map of code pieces provided by
   * {@link SourceCodeProviderTask source code providers} this task requires.
   * The returned {@link Map} associates for each source code provider the
   * <code>codeProviderName</code> to the provided code. If a source code
   * provider provides a map of code pieces rather a single string, the map
   * returned by this method associates for each code pieces:
   * <code>codeProviderName+'.'+codePieceName</code> to code piece.<br>
   * This method should only be used in the {@link Task#execute} method. <br>
   * The order of the returned map is compatible with the order of the source
   * code providers. (see {@link #AbstractRequireSourceCodeProviderTask(int)})
   * 
   * @return Returns a map of code pieces provided by
   *         {@link SourceCodeProviderTask source code providers}
   */
  protected Map<String, Object> getCodePieces() {
    final Map<String, Object> codePieces = newMap();
    final String[] codePieceNames = getSourceCodeProviderTaskNames();
    for (final String codePieceName : codePieceNames) {
      final SourceCodeProviderTask t = getSourceCodeProviderTask(codePieceName);
      final Object o = t.getSourceCode();
      if (o instanceof Map) {
        final Map<?, ?> m = (Map<?, ?>) o;
        for (final Map.Entry<?, ?> entry : m.entrySet()) {
          codePieces
              .put(codePieceName + '.' + entry.getKey(), entry.getValue());
        }
      } else {
        codePieces.put(codePieceName, o);
      }
    }
    return codePieces;
  }

  // --------------------------------------------------------------------------
  // Implementation of the RequireFactoryProviderTask interface
  // --------------------------------------------------------------------------

  public FactoryProviderTask getFactoryProviderTask() {
    return (factoryProviderTask == null)
        ? null
        : (FactoryProviderTask) factoryProviderTask.getTask();
  }

  public void setFactoryProviderTask(final TaskMap.TaskHole task) {
    if (factoryProviderTask != null) removePreviousTask(factoryProviderTask);
    factoryProviderTask = task;
    if (factoryProviderTask != null) addPreviousTask(factoryProviderTask);
  }

  // --------------------------------------------------------------------------
  // Implementation of the RequireInstanceProviderTask interface
  // --------------------------------------------------------------------------

  public InstanceProviderTask getInstanceProviderTask() {
    return (instanceProviderTask == null)
        ? null
        : (InstanceProviderTask) instanceProviderTask.getTask();
  }

  public void setInstanceProviderTask(final TaskMap.TaskHole task) {
    if (instanceProviderTask != null) removePreviousTask(instanceProviderTask);
    instanceProviderTask = task;
    if (instanceProviderTask != null) addPreviousTask(instanceProviderTask);
  }

  // --------------------------------------------------------------------------
  // Implementation of the RequiresSourceCodeProviderTask interface
  // --------------------------------------------------------------------------

  public void addSourceCodeProviderTask(final String codeProviderName,
      final TaskMap.TaskHole task) {
    if (task == null || codeProviderName == null)
      throw new NullPointerException();
    final TaskMap.TaskHole t = sourceCodeProviders.put(codeProviderName, task);
    if (t != null) removePreviousTask(t);
    addPreviousTask(task);
  }

  public void removeSourceCodeProviderTask(final String codeProviderName) {
    final TaskMap.TaskHole t = sourceCodeProviders.remove(codeProviderName);
    if (t != null) removePreviousTask(t);
  }

  public String[] getSourceCodeProviderTaskNames() {
    // use a LinkedHashSet to respect order of sourceCodeProviders.
    final List<String> names = new ArrayList<String>(sourceCodeProviders.size());
    for (final Map.Entry<String, TaskMap.TaskHole> sourceEntry : sourceCodeProviders
        .entrySet()) {
      // if sourceCodeprovider task is optional, pass it
      if (sourceEntry.getValue().isOptional()) continue;
      names.add(sourceEntry.getKey());
    }

    return names.toArray(new String[names.size()]);
  }

  public SourceCodeProviderTask getSourceCodeProviderTask(
      final String codeProviderName) {
    final TaskMap.TaskHole sourceCodeProviderTask = sourceCodeProviders
        .get(codeProviderName);
    if (sourceCodeProviderTask == null)
      throw new NoSuchElementException(codeProviderName);
    return (SourceCodeProviderTask) sourceCodeProviderTask.getTask();
  }

  // --------------------------------------------------------------------------
  // Overridden Task methods
  // --------------------------------------------------------------------------

  @Override
  protected boolean doAddDependency(final TaskHole task, final Class<?> role,
      final Object context) {
    if (super.doAddDependency(task, role, context)) return true;

    if (role == RequireInstanceProviderTask.INSTANCE_PROVIDER_TASK_ROLE)
      setInstanceProviderTask(task);
    else if (role == RequireFactoryProviderTask.FACTORY_PROVIDER_TASK_ROLE)
      setFactoryProviderTask(task);
    else if (role == RequireSourceCodeProviderTask.SOURCE_CODE_PROVIDER_TASK_ROLE)
      addSourceCodeProviderTask((String) context, task);
    else
      return false;

    return true;
  }

  @Override
  protected boolean doRemoveDependency(final TaskHole task, final Class<?> role) {
    if (super.doRemoveDependency(task, role)) return true;

    if (role == RequireInstanceProviderTask.INSTANCE_PROVIDER_TASK_ROLE)
      setInstanceProviderTask(null);
    else if (role == RequireFactoryProviderTask.FACTORY_PROVIDER_TASK_ROLE)
      setFactoryProviderTask(null);
    else if (role == RequireSourceCodeProviderTask.SOURCE_CODE_PROVIDER_TASK_ROLE)
      // TODO try to do better ...
      throw new UnsupportedOperationException();
    else
      return false;

    return true;
  }
}
